@@ -35,13 +35,16 @@ in this Software without prior written authorization from The Open Group.
 #include <stdio.h>			/* for NULL */
 #include <ctype.h>			/* for isascii() and isdigit() */
 #include <X11/Xos.h>			/* for strchr() and string routines */
-#include <X11/Xlib.h>			/* for Family contants */
+#include <X11/Xlib.h>			/* for Family constants */
 #ifdef hpux
 #include <sys/utsname.h>		/* for struct utsname */
 #endif
 #include <X11/Xauth.h>			/* for FamilyLocal */
 #include <X11/Xmu/SysUtil.h>
 
+#include <sys/stat.h>
+#include <limits.h>
+
 #if defined(UNIXCONN) || defined(LOCALCONN)
 #define UNIX_CONNECTION "unix"
 #define UNIX_CONNECTION_LENGTH 4
@@ -62,7 +65,7 @@ copystring (const char *src, int len)
     if (!src && len != 0) return NULL;
     cp = malloc (len + 1);
     if (cp) {
-	if (src) strncpy (cp, src, len);
+	if (src) memcpy (cp, src, len);
 	cp[len] = '\0';
     }
     return cp;
@@ -158,8 +161,32 @@ parse_displayname (const char *displayname,
 
     if (!host) return False;
 
-    if(strncmp (host, "/tmp/launch", 11) == 0) {
-        family = FamilyLocal;
+    {
+        /*
+         * If using launchd socket, remove the screen number from the end
+         * of $DISPLAY and check if it is a path to a socket.
+         */
+        char path[PATH_MAX];
+        struct stat sbuf;
+
+#ifdef HAVE_STRLCPY
+        strlcpy(path, displayname, sizeof(path));
+#else
+        strncpy(path, displayname, sizeof(path) - 1);
+        path[sizeof(path) - 1] = '\0';
+#endif
+        if (0 == stat(path, &sbuf) && S_ISSOCK(sbuf.st_mode)) {
+            family = FamilyLocal;
+        } else {
+            char *dot = strrchr(path, '.');
+            if (dot) {
+                *dot = '\0';
+                /* screen = atoi(dot + 1); */
+                if (0 == stat(path, &sbuf) && S_ISSOCK(sbuf.st_mode)) {
+                    family = FamilyLocal;
+                }
+            }
+        }
     }
 
     /*
